#include <uade-os-unix.h>
#include <uadeconfig.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dirent.h>
#include <libgen.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

#include "strlrep.c"

static int sharedmem_length = 65536;

static int os_debug = 0;

static int unix_soundfd;
static int unix_usingoutpipe = 0;

void uade_create_signalhandler(void (*sighandler)(int), int sigtype) {
  int signum;
  struct sigaction *sigact;
  switch (sigtype) {
  case UADE_SIGHUP:
    signum = SIGHUP;
    break;
  case UADE_SIGINT:
    signum = SIGINT;
    break;
  case UADE_SIGKILL:
    signum = SIGKILL;
    break;
  default:
    fprintf(stderr,"%s/uade: uade_create_sighandler: no such signal\n", UADE_MODULE);
    return;
  }
  sigact = calloc(1, sizeof(struct sigaction));
  sigact->sa_handler = sighandler;
  sigaction(signum, sigact, 0);
}


void uade_create_var_pid(void) {
  /* ADDED BY XIGH */
  int fd;
  char pidstr[10];
  char varname[PATH_MAX];
  if (!getenv("USER")) {
    fprintf(stderr, "uade: error: $USER not defined\n");
    return;
  }
  snprintf(varname, sizeof(varname), "/var/run/uade.%s.pid", getenv("USER"));
  sprintf(pidstr, "%d", (int) getpid());
  fd = open(varname, O_RDWR|O_CREAT|O_TRUNC, 0644);
  if (fd < 0) {
    fprintf(stderr, "could not create %s\n", varname);
  } else {
    write(fd, pidstr, strlen(pidstr));
    close(fd);
  }
}

void uade_exit(int code) {
  exit(code);
}

/* first element of args is the executable file */
/* last element of args table has to be a NULL pointer */
pid_t uade_fork_exec(char **args) {
  pid_t slavepid = fork();
  if (slavepid == 0) {
    execv(args[0], args);
    fprintf(stderr, "uade: OH SHIT! execv() has returned!\nPlease report this bug\n");
    fprintf(stderr, "uade: DON'T try to play with UADE now\n");
    abort();
  } else if (slavepid < 0) {
    fprintf(stderr, "uade: fork error\n");
    slavepid = 0;
  }
  return slavepid;
}


FILE *uade_get_file(int item) {
  char post[PATH_MAX];
  char glob[PATH_MAX];
  char temp[PATH_MAX];
  char home[PATH_MAX];
  FILE *testfile;
  if (!getenv("HOME")) {
    fprintf(stderr, "uade: warning: $HOME not defined. using '.' instead\n");
    strlcpy(home, ".", sizeof(home));
  } else {
    strlcpy(home, getenv("HOME"), sizeof(home));
  }
  post[0] = glob[0] = temp[0] = 0;
  switch (item) {
  case UADE_PATH_SCORE:
    strlcpy(post, "score", sizeof(post));
    break;
  case UADE_PATH_UAERC:
    strlcpy(post, "uaerc", sizeof(post));
    break;
  case UADE_PATH_UADE:
    strlcpy(post, "uade", sizeof(post));
    strlcpy(glob, UADECONFIG_UADE, sizeof(glob));
    break;
  case UADE_PATH_FORMATSFILE:
    strlcpy(post, "players/uadeformats", sizeof(post));
    strlcpy(glob, UADECONFIG_FORMATSNAME, sizeof(glob));
    break;
  default:
    fprintf(stderr, "uade: failed to get path of %d\n", item);
    return 0;
  }
  if (post[0]) {
    testfile = fopen(post, "r");
    if (testfile) {
      if (os_debug)
	fprintf(stderr, "uade: get file: success: path = %s\n", post);
      return testfile;
    }
    snprintf(temp, sizeof(temp), "%s/.xmms/uade/%s", home, post);
    testfile = fopen(temp, "r");
    if (testfile) {
      if (os_debug)
	fprintf(stderr, "uade: get file: success: path = %s\n", temp);
      return testfile;
    }
  }
  if (glob[0]) {
    testfile = fopen(glob, "r");
    if (testfile) {
      if (os_debug)
	fprintf(stderr, "uade: get file: success: path = %s\n", glob);
      return testfile;
    }
  }
  fprintf(stderr, "uade: failed to get path of %s\n", post);
  return 0;
}


int uade_get_path(char *path, int item, int maxlen) {
  char post[PATH_MAX];
  char glob[PATH_MAX];
  char temp[PATH_MAX];
  char home[PATH_MAX];
  FILE *testfile;
  DIR *testdir;
  post[0] = glob[0] = temp[0] = 0;
  path[0] = 0;
  if (!getenv("HOME")) {
    fprintf(stderr, "uade: warning: $HOME not defined. using '.' instead\n");
    strlcpy(home, ".", sizeof(home));
  } else {
    strlcpy(home, getenv("HOME"), sizeof(home));
  }
  switch (item) {
  case UADE_PATH_SCORE:
    strlcpy(post, "score", sizeof(post));
    strlcpy(glob, UADECONFIG_SCORENAME, sizeof(glob));
    break;
  case UADE_PATH_UAERC:
    strlcpy(post, "uaerc", sizeof(post));
    strlcpy(glob, UADECONFIG_UAERC, sizeof(glob));
    break;
  case UADE_PATH_UADE:
    strlcpy(post, "uade", sizeof(post));
    /* test if 'uade' is a directory */
    testdir = opendir(post);
    if(testdir) {
      /* 'uade' is a directory => don't use it */
      closedir(testdir);
      post[0] = 0;
    }
    strlcpy(glob, UADECONFIG_UADE, sizeof(glob));
    break;
  case UADE_PATH_PLAYERDIR:
    strlcpy(glob, "players/", sizeof(glob));
    testdir = opendir(glob);
    if(testdir) {
      closedir(testdir);
      strlcpy(path, glob, maxlen);
      strlcat(glob, "uadeformats", sizeof(glob));
      testfile = fopen(glob, "r");
      if(testfile) {
	fclose(testfile); testfile = 0;
	if(os_debug)
	  fprintf(stderr, "success: path = %s\n", path);
	return 1;
      }
    }
    snprintf(glob, sizeof(glob), "%s/.xmms/uade/players/", home);
    testdir = opendir(glob);
    if(testdir) {
      closedir(testdir);
      strlcpy(path, glob, maxlen);
      if(os_debug)
	fprintf(stderr, "success: path = %s\n", path);
      return 1;
    }
    snprintf(glob, sizeof(glob), "%splayers/", UADECONFIG_DATADIR);
    testdir = opendir(glob);
    if(testdir) {
      closedir(testdir);
      strlcpy(path, glob, maxlen);
      if(os_debug)
	fprintf(stderr, "success: path = %s\n", path);
      return 1;
    }
    fprintf(stderr, "uade: couldn't get uade playerdir path\n");
    return 0;
  case UADE_PATH_FORMATSFILE:
    strlcpy(post, "players/uadeformats", sizeof(post));
    strlcpy(glob, UADECONFIG_FORMATSNAME, sizeof(glob));
    break;
  default:
    fprintf(stderr, "uade: failed to get path of %d\n", item);
    return 0;
  }
  if(post[0]) {
    testfile = fopen(post, "r");
    if(testfile) {
      fclose(testfile); testfile = 0;
      strlcpy(path, post, maxlen);
      if(os_debug)
	fprintf(stderr, "uade: success: path = %s\n", path);
      return 1;
    }
    snprintf(temp, sizeof(temp), "%s/.xmms/uade/%s", home, post);
    testfile = fopen(temp, "r");
    if(testfile) {
      fclose(testfile); testfile = 0;
      strlcpy(path, temp, maxlen);
      if(os_debug)
	fprintf(stderr, "uade: success: path = %s\n", path);
      return 1;
    }
  }
  if(glob[0]) {
    testfile = fopen(glob, "r");
    if(os_debug)
      fprintf(stderr, "uade: getpath: trying glob = %s\n", glob);
    if(testfile) {
      fclose(testfile); testfile = 0;
      strlcpy(path, glob, maxlen);
      if(os_debug)
	fprintf(stderr, "uade: success: path = %s\n", path);
      return 1;
    }
  }
  fprintf(stderr,"uade: failed to get path of %s\n", post);
  return 0;
}


int uade_get_postfix(char *dst, char *filename, int maxlen) {
  char *baseptr;
  if (maxlen == 0)
    return 0;
  baseptr = strrchr(filename, '/');
  if (baseptr) {
    baseptr++;
  } else {
    baseptr = filename;
  }
  baseptr = strrchr(baseptr, (int) '.');
  if (!baseptr) {
    return 0;
  }
  baseptr++;
  if (((int) strlen(baseptr)) >= maxlen) {
    return 0;
  }
  strcpy(dst, baseptr); /* will not overflow */
  return 1;
}


int uade_get_prefix(char *dst, char *filename, int maxlen) {
  int i, ret;
  char *baseptr;
  if (maxlen == 0)
    return 0;
  baseptr = strrchr(filename, '/');
  if (baseptr) {
    baseptr++;
  } else {
    baseptr = filename;
  }
  if (!strchr(baseptr, (int) '.')) {
    return 0;
  }
  ret = 0;
  for (i=0;;i++) {
    if (i >= maxlen)
      break;
    if (baseptr[i] == '.') {
      ret = 1;
      break;
    }
    dst[i] = baseptr[i];
  }
  dst[i] = 0;
  return ret;
}


int uade_get_temp_name(char *tempname, int maxlen) {
  int fd;
  char dir[PATH_MAX];
  char tmpdir[PATH_MAX];
  char user[1024] = "";

  if (!maxlen) {
    fprintf(stderr, "uade: ERROR: uade_get_temp_name: maxlen == 0\n");
    return 0;
  }

  /* find out temp directory */
  if (getenv("TEMP")) {
    strlcpy(tmpdir, getenv("TEMP"), sizeof(tmpdir));
  } else if (getenv("TMP")) {
    strlcpy(tmpdir, getenv("TMP"), sizeof(tmpdir));
  } else {
    strlcpy(tmpdir, "/tmp", sizeof(tmpdir));
  }

  if (getlogin()) {
    strcpy(user, getlogin());
  } else {
    sprintf(user, "%d", (int) getuid());
  }

  snprintf(dir, sizeof(dir), "%s/%s.uade", tmpdir, user);

  if (mkdir(dir, S_IRUSR | S_IWUSR | S_IXUSR)) {
    if (chmod(dir, S_IRUSR | S_IWUSR | S_IXUSR)) {
      fprintf(stderr, "uade: couldn't create directory for tmpfile\n");
      tempname[0] = 0;
      return 0;
    }
  }
  snprintf(tempname, maxlen, "%s/mmap.%d.XXXXXX", dir, (int) getpid());
  fd = mkstemp(tempname);
  if (fd < 0) {
    fprintf(stderr, "uade: couldn't generate tmp filename!\n");
    tempname[0] = 0;
    return 0;
  }
  close(fd);
  remove(tempname);
  return 1;
}


int uade_init_mmap_file(const char *mapfilename) {
  char zerobuf[256];
  FILE *sharedmemFILE;
  int i;
  if(!(sharedmemFILE = fopen(mapfilename,"w+"))) {
    fprintf(stderr,"uade: can not create mmap file (fopen)\n");
    return 0;
  }
  memset(zerobuf, 0, 256);
  for(i=0; i<sharedmem_length; i+=256) {
    fwrite(zerobuf, 1, 256, sharedmemFILE);
  }
  fclose(sharedmemFILE);
  return 1;
}


int uade_is_regular_file(const char *filename) {
  struct stat s;
  if (stat(filename, &s) == 0) {
    if (S_ISREG(s.st_mode)) {
      return 1;
    }
  }
  return 0;
}

/* THIS IS UGLY, AND HAS A FEW BUGS. NO EXPLOITS AFAIK, BUT STILL NEEDS
   REWRITE. */
FILE *uade_open_amiga_file(char *amigafilename) {
  FILE *file;
  char *separator;
  char *name;         /* points to file without path */
  char filename[PATH_MAX];
  char path[PATH_MAX];    /* contains dirpath to the file */
  char loadname[PATH_MAX];
  DIR *dir;
  struct dirent *direntry;
  int maxsize;

  if(strlen(amigafilename) >= 256) {
    fprintf(stderr, "uade: WARNING: amiga tried to open a very long filename\nPlease REPORT THIS!\n");
  }
  maxsize = (sizeof(filename) < 256) ? sizeof(filename) : 256;
  strlcpy(filename, amigafilename, maxsize);

  if(!strncasecmp(filename, "S:", 2)) {
    filename[0] = 'S';
    uade_get_path(path, UADE_PATH_PLAYERDIR, sizeof(path));
    strlcat(path, "/", sizeof(path));
    strlcat(path, filename, sizeof(path));
    strlcpy(filename, path, sizeof(filename));
  }

  if(!strncasecmp(filename, "ENV:", 4))
    memcpy(filename, "ENV:", 4);

  if ((file = fopen(filename,"rb"))) {
    return file;
  }

  /* case: filename = VOLUME:PATH1[/FILE] */
  if((separator = strchr(filename, (int) ':'))) {
    uade_get_path(path, UADE_PATH_PLAYERDIR, sizeof(path));
    if((name = strrchr(separator + 1, (int) '/'))) {
      *name = 0;
      strlcat(path, filename, sizeof(path));
      strlcat(path, "/", sizeof(path));
      *name = '/';
      name++;
    } else {
      name = separator + 1;
    }
    // fprintf(stderr, "uade: using load path = %s name = %s\n", path, name);
    goto jumpover;
  }

  if((separator = strrchr(filename,(int) '/'))) {
    name = separator + 1;
    *separator = 0;
    snprintf(path, sizeof(path), "%s/", filename);
    *separator = '/';
  } else {
    name = filename;
    strlcpy(path,"./", sizeof(path));
  }

 jumpover:

  if((dir = opendir(path)) == 0) {
    path[strlen(path)-1] = 0;
    if((separator = strrchr(path, (int) '/'))) {
      if(isupper((int) separator[1])) {
	separator[1] = (char) tolower((int) separator[1]);
      } else if(islower((int) separator[1])) {
	separator[1] = (char) toupper((int) separator[1]);
      }
      strlcat(path, "/", sizeof(path));
      if((dir = opendir(path))==0) {
	fprintf(stderr, "uade: open file: can't open module directory '%s' for scanning!\n", path);
	return 0;
      }
    } else {
      fprintf(stderr, "uade: open file: Can't open module directory '%s' for scanning!\n", path);
      return 0;
    }
  }

  /* fprintf(stderr,"name = '%s' path = '%s'\n",name,path); */
  while((direntry = readdir(dir))) {
    /* fprintf(stderr,"dirent: %s\n",direntry->d_name); */
    if(strcasecmp(name, direntry->d_name)==0) {
      snprintf(loadname, sizeof(loadname), "%s%s", path, direntry->d_name);
      if((file = fopen(loadname, "rb"))) break;
    }
  }
  closedir(dir);

  if(!file)
    fprintf(stderr, "uade: couldn't load file '%s'\n", filename);

  return file;
}


void uade_init_outpipe(char *parameter) {
  unix_usingoutpipe = 1;
  unix_soundfd = atoi(parameter);
}


void *uade_mmap_file(const char *filename) {
  void *mmapptr;
  int fd;
  fd = open(filename,O_RDWR);
  if(fd < 0) {
    fprintf(stderr,"uade: can not open sharedmem file!\n");
    return 0;
  }
  mmapptr = mmap(0, sharedmem_length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
  if(mmapptr == MAP_FAILED) {
    fprintf(stderr,"uade: can not mmap sharedmem file!\n");
    return 0;
  }
  return mmapptr;
}


int uade_send_signal(pid_t pid, int sigtype) {
  int signum;
  switch(sigtype) {
  case UADE_SIGALIVE:
    signum = 0;
    break;
  case UADE_SIGHUP:
    signum = SIGHUP;
    break;
  case UADE_SIGINT:
    signum = SIGINT;
    break;
  case UADE_SIGKILL:
    signum = SIGKILL;
    break;
  default:
    fprintf(stderr,"%s/uade: uade_send_signal: unknown signal %d\n", UADE_MODULE, sigtype);
    return -1;
  }
  if(pid > 0)
    return kill(pid, signum);
  else
    return -1;
}


void uade_sleep(int seconds) {
  sleep(seconds);
}


char *uade_strdup_basename(char *path) {
  char *copy;
  char *bn;
  if (!path)
    return 0;
  copy = strdup(path);
  if (!copy) {
    return 0;
  }
  bn = basename(copy);
  if (!bn) {
    goto error;
  }
  bn = strdup(bn);
 error:
  free(copy);
  return bn;
}


void uade_usleep(int microseconds) {
  usleep(microseconds);
}


/* returns number of milliseconds from the last call
   exception: on the first time this returns zero    */
int uade_timer_poll(void) {
  static struct timeval otv = {0,0};
  struct timeval tv;
  int ret = 0;
  gettimeofday(&tv, 0);
  if(otv.tv_sec != 0 || otv.tv_usec != 0) {
    ret = (tv.tv_sec-otv.tv_sec)*1000 + (tv.tv_usec-otv.tv_usec)/1000;
  }
  otv = tv;
  return ret;
}

char *uade_version(void) {
  return strdup(UADEVERSIONSTRING);
}

void uade_write_to_outpipe(void *ptr, int size) {
  int ret, written;
  if (unix_usingoutpipe) {
    written = 0;
    while (written < size) {
      ret = write(unix_soundfd, ptr, size);
      if (ret > 0) {
	written += ret;
      } else if (ret == 0) {
	break;
      } else {
	if (errno != EINTR) {
	  break;
	}
      }
    }
  }
}
