/* Copyright 2003 Heikki Orsila <heikki.orsila@ee.tut.fi>
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/poll.h>
#include <errno.h>
#include <limits.h>
#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#include "unix-shell.h"

#include "uade-os.h"
#include "../osdep/strl.c"

struct songnode_t {
  struct songnode_t *next;
  char filename[PATH_MAX];
};

static struct songnode_t *SongList = 0;

static void (*us_interrupt)(void) = 0;

/* catch CTRL-C and cause REBOOT */
static void us_sigint_handler(int signum) {
  signum = signum;
  if (us_interrupt) {
    us_interrupt();
  } else {
    fprintf(stderr, "uade: non-initialized sigint interrupt\n");
  }
}

static void us_setup(void (*interrupt)(void)) {
  us_interrupt = interrupt;
  /* create CTRL-C break for cmdline tool */
  uade_create_signalhandler(us_sigint_handler, UADE_SIGINT);
}

static void us_enqueue_song(char *file) {
  struct songnode_t *s = SongList;
  struct songnode_t *new;

  /* allow queuing of zero length string */
  if (uade_is_regular_file(file) || strlen(file) == 0) {
    new = malloc(sizeof(struct songnode_t));
    new->next = 0;
    strlcpy(new->filename, file, sizeof(new->filename));
    if (s) {
      while(s->next) s = s->next;
      s->next = new;
    } else {
      SongList = new;
    }
  }
}

static int us_song_list_empty(void) {
  return SongList ? 0 : 1;
}

static void us_get_next(char *dst, int maxsize) {
  char playername[PATH_MAX];
  char tempname[PATH_MAX];

  if (!SongList)
    return;

  strlcpy(dst, SongList->filename, maxsize);
  SongList = SongList->next;
}

static int us_interaction(struct uade_command *cmd) {
  struct pollfd ifd;
  int ret;
  char buf[256];

  memset(&ifd, 0, sizeof(ifd));
  ifd.fd = 0;
  ifd.events = POLLIN;
  ret = poll(&ifd, 1, 0);
  if (ret > 0) {
    if (ifd.revents & POLLIN) {
      ret = read(ifd.fd, buf, 2);
      if (ret < 0) {
	if (errno != EINTR) {
	  goto no_interaction;
	}
      } else if (ret == 0) {
	goto no_interaction;
      } else {
	if (buf[1] == '\n') {
	  switch (buf[0]) {
	  case '\n':
	    break;

	  case 'n':
	    cmd->type = UADE_REBOOT;
	    return 1;

	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
	    cmd->type = UADE_SETSUBSONG;
	    cmd->ret = malloc(sizeof(int));
	    if (!cmd->ret) {
	      goto no_interaction;
	    }
	    *((int *) cmd->ret) = (int) (buf[0] - '0');
	    return 1;

	  default:
	    fprintf(stderr, "uade: unknown command for interaction\n");
	    break;
	  }
	}
      }
    }

  } else if (ret < 0) {
    if (errno != EINTR) {
      goto no_interaction;
    }
  }

  return 0;

 no_interaction:
  cmd->type = UADE_NO_INTERACTION;
  return 1;
}

void us_functions(struct uade_local *local) {
  local->setup = us_setup;
  local->enqueue_song = us_enqueue_song;
  local->get_next = us_get_next;
  local->interaction = us_interaction;
  local->song_list_empty = us_song_list_empty;
};
